@@ -3,6 +3,7 @@ from re import sub
 
 from flask import Blueprint, render_template, url_for, abort, g
 from flask.ext.admin import babel
+from flask.ext.admin._compat import with_metaclass
 from flask.ext.admin import helpers as h
 
 
@@ -34,14 +35,20 @@ def expose_plugview(url='/'):
         .. versionadded:: 1.0.4
     """
     def wrap(v):
-        return expose(url, v.methods)(v.as_view(v.__name__))
+        handler = expose(url, v.methods)
+
+        if hasattr(v, 'as_view'):
+            return handler(v.as_view(v.__name__))
+        else:
+            return handler(v)
+
     return wrap
 
 
 # Base views
 def _wrap_view(f):
     @wraps(f)
-    def inner(self, **kwargs):
+    def inner(self, *args, **kwargs):
         # Store current admin view
         h.set_current_view(self)
 
@@ -50,7 +57,7 @@ def _wrap_view(f):
         if abort is not None:
             return abort
 
-        return f(self, **kwargs)
+        return f(self, *args, **kwargs)
 
     return inner
 
@@ -84,7 +91,11 @@ class AdminViewMeta(type):
                 setattr(cls, p, _wrap_view(attr))
 
 
-class BaseView(object):
+class BaseViewClass(object):
+    pass
+
+
+class BaseView(with_metaclass(AdminViewMeta, BaseViewClass)):
     """
         Base administrative view.
 
@@ -95,8 +106,6 @@ class BaseView(object):
                 def index(self):
                     return 'Hello World!'
     """
-    __metaclass__ = AdminViewMeta
-
     @property
     def _template_args(self):
         """
@@ -128,7 +137,8 @@ class BaseView(object):
 
         return args
 
-    def __init__(self, name=None, category=None, endpoint=None, url=None, static_folder=None, static_url_path=None):
+    def __init__(self, name=None, category=None, endpoint=None, url=None,
+                 static_folder=None, static_url_path=None):
         """
             Constructor.
 
@@ -148,6 +158,9 @@ class BaseView(object):
                 and '/admin/' prefix won't be applied.
             :param static_url_path:
                 Static URL Path. If provided, this specifies the path to the static url directory.
+            :param debug:
+                Optional debug flag. If set to `True`, will rethrow exceptions in some cases, so Werkzeug
+                debugger can catch them.
         """
         self.name = name
         self.category = category
@@ -162,7 +175,7 @@ class BaseView(object):
 
         # Default view
         if self._default_view is None:
-            raise Exception('Attempted to instantiate admin view %s without default view' % self.__class__.__name__)
+            raise Exception(u'Attempted to instantiate admin view %s without default view' % self.__class__.__name__)
 
     def create_blueprint(self, admin):
         """
@@ -249,6 +262,17 @@ class BaseView(object):
         """
         return sub(r'(?<=.)([A-Z])', r' \1', name)
 
+    def is_visible(self):
+        """
+            Override this method if you want dynamically hide or show administrative views
+            from Flask-Admin menu structure
+
+            By default, item is visible in menu.
+
+            Please note that item should be both visible and accessible to be displayed in menu.
+        """
+        return True
+
     def is_accessible(self):
         """
             Override this method to add permission checks.
@@ -273,7 +297,14 @@ class BaseView(object):
                 View function arguments
         """
         if not self.is_accessible():
-            return abort(404)
+            return abort(403)
+
+    @property
+    def _debug(self):
+        if not self.admin or not self.admin.app:
+            return False
+
+        return self.admin.app.debug
 
 
 class AdminIndexView(BaseView):
@@ -348,6 +379,12 @@ class MenuItem(object):
 
         return view.url in self._children_urls
 
+    def is_visible(self):
+        if self._view is None:
+            return False
+
+        return self._view.is_visible()
+
     def is_accessible(self):
         if self._view is None:
             return False
@@ -358,12 +395,12 @@ class MenuItem(object):
         return self._view is None
 
     def get_children(self):
-        return [c for c in self._children if c.is_accessible()]
+        return [c for c in self._children if c.is_accessible() and c.is_visible()]
 
 
 class MenuLink(object):
     """
-        Menu additional links hierarchy.
+        Additional menu links.
     """
     def __init__(self, name, url=None, endpoint=None):
         self.name = name
@@ -373,6 +410,9 @@ class MenuLink(object):
     def get_url(self):
         return self.url or url_for(self.endpoint)
 
+    def is_visible(self):
+        return True
+
     def is_accessible(self):
         return True
 
@@ -436,11 +476,8 @@ class Admin(object):
         # Add predefined index view
         self.add_view(self.index_view)
 
-        # Localizations
-        self.locale_selector_func = None
-
         # Register with application
-        if app:
+        if app is not None:
             self._init_extension()
 
     def add_view(self, view):
@@ -467,37 +504,6 @@ class Admin(object):
         """
         self._menu_links.append(link)
 
-    def locale_selector(self, f):
-        """
-            Installs a locale selector for the current ``Admin`` instance.
-
-            Example::
-
-                def admin_locale_selector():
-                    return request.args.get('lang', 'en')
-
-                admin = Admin(app)
-                admin.locale_selector(admin_locale_selector)
-
-            It is also possible to use the ``@admin`` decorator::
-
-                admin = Admin(app)
-
-                @admin.locale_selector
-                def admin_locale_selector():
-                    return request.args.get('lang', 'en')
-
-            Or by subclassing the ``Admin``::
-
-                class MyAdmin(Admin):
-                    def locale_selector(self):
-                        return request.args.get('lang', 'en')
-        """
-        if self.locale_selector_func is not None:
-            raise Exception('Can not add locale_selector second time.')
-
-        self.locale_selector_func = f
-
     def _add_view_to_menu(self, view):
         """
             Add a view to the menu tree
@@ -541,12 +547,12 @@ class Admin(object):
 
         for p in admins:
             if p.endpoint == self.endpoint:
-                raise Exception('Cannot have two Admin() instances with same'
-                                ' endpoint name.')
+                raise Exception(u'Cannot have two Admin() instances with same'
+                                u' endpoint name.')
 
             if p.url == self.url and p.subdomain == self.subdomain:
-                raise Exception('Cannot assign two Admin() instances with same'
-                                ' URL and subdomain to the same application.')
+                raise Exception(u'Cannot assign two Admin() instances with same'
+                                u' URL and subdomain to the same application.')
 
         admins.append(self)
         self.app.extensions['admin'] = admins
