@@ -1,19 +1,29 @@
 import logging
 
-from flask import flash
+from flask import request, flash, abort, Response
 
+from flask.ext.admin import expose
 from flask.ext.admin.babel import gettext, ngettext, lazy_gettext
 from flask.ext.admin.model import BaseModelView
+from flask.ext.admin._compat import iteritems, string_types
 
 import mongoengine
+import gridfs
+from mongoengine.connection import get_db
 from bson.objectid import ObjectId
 
 from flask.ext.admin.actions import action
-from flask.ext.admin.form import BaseForm
 from .filters import FilterConverter, BaseMongoEngineFilter
-from .form import model_form, CustomModelConverter
+from .form import get_form, CustomModelConverter
 from .typefmt import DEFAULT_FORMATTERS
 from .tools import parse_like_term
+from .helpers import format_error
+from .ajax import process_ajax_references, create_ajax_loader
+from .subdoc import convert_subdocuments
+
+
+# Set up logger
+log = logging.getLogger("flask-admin.mongo")
 
 
 SORTABLE_FIELDS = set((
@@ -29,7 +39,7 @@ SORTABLE_FIELDS = set((
     mongoengine.EmailField,
     mongoengine.UUIDField,
     mongoengine.URLField
-    ))
+))
 
 
 class ModelView(BaseModelView):
@@ -61,6 +71,9 @@ class ModelView(BaseModelView):
         Model form conversion class. Use this to implement custom
         field conversion logic.
 
+        Custom class should be derived from the
+        `flask.ext.admin.contrib.mongoengine.form.CustomModelConverter`.
+
         For example::
 
             class MyModelConverter(AdminModelConverter):
@@ -83,6 +96,92 @@ class ModelView(BaseModelView):
         Customized type formatters for MongoEngine backend
     """
 
+    allowed_search_types = (mongoengine.StringField,
+                            mongoengine.URLField,
+                            mongoengine.EmailField)
+    """
+        List of allowed search field types.
+    """
+
+    form_subdocuments = None
+    """
+        Subdocument configuration options.
+
+        This field accepts dictionary, where key is field name and value is either dictionary or instance of the
+        `flask.ext.admin.contrib.EmbeddedForm`.
+
+        Consider following example::
+
+            class Comment(db.EmbeddedDocument):
+                name = db.StringField(max_length=20, required=True)
+                value = db.StringField(max_length=20)
+
+            class Post(db.Document):
+                text = db.StringField(max_length=30)
+                data = db.EmbeddedDocumentField(Comment)
+
+            class MyAdmin(ModelView):
+                form_subdocuments = {
+                    'data': {
+                        'form_columns': ('name',)
+                    }
+                }
+
+        In this example, `Post` model has child `Comment` subdocument. When generating form for `Comment` embedded
+        document, Flask-Admin will only create `name` field.
+
+        It is also possible to use class-based embedded document configuration::
+
+            class CommentEmbed(EmbeddedForm):
+                form_columns = ('name',)
+
+            class MyAdmin(ModelView):
+                form_subdocuments = {
+                    'data': CommentEmbed()
+                }
+
+        Arbitrary depth nesting is supported::
+
+            class SomeEmbed(EmbeddedForm):
+                form_excluded_columns = ('test',)
+
+            class CommentEmbed(EmbeddedForm):
+                form_columns = ('name',)
+                form_subdocuments = {
+                    'inner': SomeEmbed()
+                }
+
+            class MyAdmin(ModelView):
+                form_subdocuments = {
+                    'data': CommentEmbed()
+                }
+
+        There's also support for forms embedded into `ListField`. All you have
+        to do is to create nested rule with `None` as a name. Even though it
+        is slightly confusing, but that's how Flask-MongoEngine creates
+        form fields embedded into ListField::
+
+            class Comment(db.EmbeddedDocument):
+                name = db.StringField(max_length=20, required=True)
+                value = db.StringField(max_length=20)
+
+            class Post(db.Document):
+                text = db.StringField(max_length=30)
+                data = db.ListField(db.EmbeddedDocumentField(Comment))
+
+            class MyAdmin(ModelView):
+                form_subdocuments = {
+                    'data': {
+                        'form_subdocuments': {
+                            None: {
+                                'form_columns': ('name',)
+                            }
+                        }
+
+                    }
+                }
+    """
+
     def __init__(self, model, name=None,
                  category=None, endpoint=None, url=None):
         """
@@ -105,6 +204,31 @@ class ModelView(BaseModelView):
 
         self._primary_key = self.scaffold_pk()
 
+    def _refresh_cache(self):
+        """
+            Refresh cache.
+        """
+        # Process subdocuments
+        if self.form_subdocuments is None:
+            self.form_subdocuments = {}
+
+        self._form_subdocuments = convert_subdocuments(self.form_subdocuments)
+
+        # Cache other properties
+        super(ModelView, self)._refresh_cache()
+
+    def _process_ajax_references(self):
+        """
+            AJAX endpoint is exposed by top-level admin view class, but
+            subdocuments might have AJAX references too.
+
+            This method will recursively go over subdocument configuration
+            and will precompute AJAX references for them ensuring that
+            subdocuments can also use AJAX to populate their ReferenceFields.
+        """
+        references = super(ModelView, self)._process_ajax_references()
+        return process_ajax_references(references, self)
+
     def _get_model_fields(self, model=None):
         """
             Inspect model and return list of model fields
@@ -115,7 +239,7 @@ class ModelView(BaseModelView):
         if model is None:
             model = self.model
 
-        return sorted(model._fields.iteritems(), key=lambda n: n[1].creation_counter)
+        return sorted(iteritems(model._fields), key=lambda n: n[1].creation_counter)
 
     def scaffold_pk(self):
         # MongoEngine models have predefined 'id' as a key
@@ -171,7 +295,7 @@ class ModelView(BaseModelView):
         """
         if self.column_searchable_list:
             for p in self.column_searchable_list:
-                if isinstance(p, basestring):
+                if isinstance(p, string_types):
                     p = self.model._fields.get(p)
 
                 if p is None:
@@ -180,7 +304,7 @@ class ModelView(BaseModelView):
                 field_type = type(p)
 
                 # Check type
-                if (field_type != mongoengine.StringField):
+                if (field_type not in self.allowed_search_types):
                         raise Exception('Can only search on text columns. ' +
                                         'Failed to setup search for "%s"' % p)
 
@@ -195,7 +319,7 @@ class ModelView(BaseModelView):
             :param name:
                 Either field name or field instance
         """
-        if isinstance(name, basestring):
+        if isinstance(name, string_types):
             attr = self.model._fields.get(name)
         else:
             attr = name
@@ -206,7 +330,7 @@ class ModelView(BaseModelView):
         # Find name
         visible_name = None
 
-        if not isinstance(name, basestring):
+        if not isinstance(name, string_types):
             visible_name = self.get_column_name(attr.name)
 
         if not visible_name:
@@ -230,16 +354,23 @@ class ModelView(BaseModelView):
         return isinstance(filter, BaseMongoEngineFilter)
 
     def scaffold_form(self):
-        # TODO: Fix base_class
-        form_class = model_form(self.model,
-                        base_class=BaseForm,
-                        only=self.form_columns,
-                        exclude=self.form_excluded_columns,
-                        field_args=self.form_args,
-                        converter=self.model_form_converter())
+        """
+            Create form from the model.
+        """
+        form_class = get_form(self.model,
+                              self.model_form_converter(self),
+                              base_class=self.form_base_class,
+                              only=self.form_columns,
+                              exclude=self.form_excluded_columns,
+                              field_args=self.form_args,
+                              extra_fields=self.form_extra_fields)
 
         return form_class
 
+    # AJAX foreignkey support
+    def _create_ajax_loader(self, name, opts):
+        return create_ajax_loader(self.model, name, name, opts)
+
     def get_query(self):
         """
         Returns the QuerySet for this view.  By default, it returns all the
@@ -300,6 +431,11 @@ class ModelView(BaseModelView):
         # Sorting
         if sort_column:
             query = query.order_by('%s%s' % ('-' if sort_desc else '', sort_column))
+        else:
+            order = self._get_default_order()
+
+            if order:
+                query = query.order_by('%s%s' % ('-' if order[1] else '', order[0]))
 
         # Pagination
         if page is not None:
@@ -319,7 +455,13 @@ class ModelView(BaseModelView):
             :param id:
                 Model ID
         """
-        return self.get_query().filter(pk=id).first()
+        try:
+            return self.get_query().filter(pk=id).first()
+        except mongoengine.ValidationError as ex:
+            flash(gettext('Failed to get model. %(error)s',
+                          error=format_error(ex)),
+                  'error')
+            return None
 
     def create_model(self, form):
         """
@@ -331,14 +473,21 @@ class ModelView(BaseModelView):
         try:
             model = self.model()
             form.populate_obj(model)
-            self.on_model_change(form, model)
+            self._on_model_change(form, model, True)
             model.save()
-            return True
-        except Exception, ex:
-            flash(gettext('Failed to create model. %(error)s', error=str(ex)),
-                'error')
-            logging.exception('Failed to create model')
+        except Exception as ex:
+            if self._debug:
+                raise
+
+            flash(gettext('Failed to create model. %(error)s',
+                          error=format_error(ex)),
+                  'error')
+            log.exception('Failed to create model')
             return False
+        else:
+            self.after_model_change(form, model, True)
+
+        return True
 
     def update_model(self, form, model):
         """
@@ -351,14 +500,21 @@ class ModelView(BaseModelView):
         """
         try:
             form.populate_obj(model)
-            self.on_model_change(form, model)
+            self._on_model_change(form, model, False)
             model.save()
-            return True
-        except Exception, ex:
-            flash(gettext('Failed to update model. %(error)s', error=str(ex)),
-                'error')
-            logging.exception('Failed to update model')
+        except Exception as ex:
+            if self._debug:
+                raise
+
+            flash(gettext('Failed to update model. %(error)s',
+                          error=format_error(ex)),
+                  'error')
+            log.exception('Failed to update model')
             return False
+        else:
+            self.after_model_change(form, model, False)
+
+        return True
 
     def delete_model(self, model):
         """
@@ -371,12 +527,38 @@ class ModelView(BaseModelView):
             self.on_model_delete(model)
             model.delete()
             return True
-        except Exception, ex:
-            flash(gettext('Failed to delete model. %(error)s', error=str(ex)),
-                'error')
-            logging.exception('Failed to delete model')
+        except Exception as ex:
+            if self._debug:
+                raise
+
+            flash(gettext('Failed to delete model. %(error)s',
+                          error=format_error(ex)),
+                  'error')
+            log.exception('Failed to delete model')
             return False
 
+    # FileField access API
+    @expose('/api/file/')
+    def api_file_view(self):
+        pk = request.args.get('id')
+        coll = request.args.get('coll')
+        db = request.args.get('db', 'default')
+
+        if not pk or not coll or not db:
+            abort(404)
+
+        fs = gridfs.GridFS(get_db(db), coll)
+
+        data = fs.get(ObjectId(pk))
+        if not data:
+            abort(404)
+
+        return Response(data.read(),
+                        content_type=data.content_type,
+                        headers={
+                            'Content-Length': data.length
+                        })
+
     # Default model actions
     def is_action_allowed(self, name):
         # Check delete action permission
@@ -394,13 +576,15 @@ class ModelView(BaseModelView):
 
             all_ids = [ObjectId(pk) for pk in ids]
             for obj in self.get_query().in_bulk(all_ids).values():
-                obj.delete()
-                count += 1
+                count += self.delete_model(obj)
 
             flash(ngettext('Model was successfully deleted.',
                            '%(count)s models were successfully deleted.',
                            count,
                            count=count))
-        except Exception, ex:
+        except Exception as ex:
+            if self._debug:
+                raise
+
             flash(gettext('Failed to delete models. %(error)s', error=str(ex)),
-                'error')
+                  'error')
